using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;
using System.Threading;
using IFXplus.SmartStorage.Settings;
using IFXplus.SmartStorage.Utils;

namespace IFXplus.SmartStorage.Providers
{
    internal class FileSystemStorageProvider : IStorageProvider
    {

        #region Const

        private static readonly char[] SplitPathChars = ("/.(=,").ToCharArray();
        private static readonly string FileExtension = ".bin";

        #endregion

        #region Privates

        private readonly StorageProviderReference _reference = null;
        private readonly FileSystemStorageProviderSettings _settings = null;

        #endregion

        #region Constructor

        public FileSystemStorageProvider(StorageProviderReference reference)
        {
            if (reference == null)
                throw new ArgumentNullException("reference");
            if (string.IsNullOrEmpty(reference.Id))
                throw new ArgumentException("reference.Id cannot be null or empty", "reference");
            _reference = reference;
            
            _settings = StorageSettings.GetInstance().GetStorageProviderSettings<FileSystemStorageProviderSettings>(reference.Id);
            if (_settings == null)
                throw new StorageSettingsException(string.Format("{0} provider settings not found", reference));

            if (!Directory.Exists(Settings.Path))
                Directory.CreateDirectory(Settings.Path);

            _DebugMode = Settings.DebugMode;
            _LogFilePath = Settings.ScavengingLogFile;
            FileInfo logfile = new FileInfo(_LogFilePath);
            if (!logfile.Directory.Exists)
                Directory.CreateDirectory(logfile.DirectoryName);

            StartScavenging();
        }

        #endregion

        #region Properties

        private FileSystemStorageProviderSettings Settings
        {
            get { return _settings; }
        }

        #endregion

        #region IStorageProvider Members

        public StorageProviderReference Reference
        {
            get { return _reference; }
        }

        public void Insert(StorageEntry entry)
        {
            try
            {
                InsertItem(GetFileInfo(entry.Key), entry);                
            }
            catch (Exception ex)
            {
                throw new StorageException(string.Format("Exception in FileSystemStorageProvider.Insert({0})", entry), ex);
            }            
        }

        public void Remove(string key)
        {
            try
            {
                RemoveItem(GetFileInfo(key));
            }
            catch (Exception ex)
            {
                throw new StorageException(string.Format("Exception in FileSystemStorageProvider.Remove({0})", key), ex);
            }
        }

        public StorageEntry Get(string key)
        {
            try
            {
                return GetItem(GetFileInfo(key));
            }
            catch (Exception ex)
            {
                throw new StorageException(string.Format("Exception in FileSystemStorageProvider.Get({0})", key), ex);
            }
        }

        #endregion

        #region IO Utils

        private static readonly int ReaderLockTimeout = Timeout.Infinite;
        private static readonly int WriterLockTimeout = Timeout.Infinite;
        private static readonly ReaderWriterLock _sync = new ReaderWriterLock();
        
        private FileInfo GetFileInfo(string key)
        {            
            string path = Settings.Path;
            Regex re = new Regex(@"([^A-Za-z0-9_\-.,=])");
            string[] subFolders = key.Split(SplitPathChars, StringSplitOptions.RemoveEmptyEntries);
            foreach (string f in subFolders)
                path = Path.Combine(path, re.Replace(f, ""));
            return new FileInfo(path + FileExtension);
        }

        private static void InsertItem(FileInfo file, StorageEntry entry)
        {
            _sync.AcquireWriterLock(WriterLockTimeout);
            try
            {                
                if (file.Exists)
                    file.Delete();
                else
                {
                    if (!Directory.Exists(file.DirectoryName))
                        Directory.CreateDirectory(file.DirectoryName);
                }
                BinarySerializer.Serialize(file, entry);
            }
            catch (Exception ex)
            {
                throw new StorageException(string.Format("Exception in FileSystemStorageProvider.InsertItem({0})", file.FullName), ex);
            }
            finally
            {
                _sync.ReleaseWriterLock();
            }
        }

        private static StorageEntry GetItem(FileInfo file)
        {
            _sync.AcquireReaderLock(ReaderLockTimeout);
            try
            {
                return (file.Exists) ? (StorageEntry) BinarySerializer.Deserialize(file) : null;
            }
            catch (Exception ex)
            {
                throw new StorageException(string.Format("Exception in FileSystemStorageProvider.GetStorageEntry({0})", file.FullName), ex);
            }
            finally
            {
                _sync.ReleaseReaderLock();
            }
        }

        private static void RemoveItem(FileSystemInfo file)
        {
            _sync.AcquireWriterLock(WriterLockTimeout);
            try
            {
                if (file.Exists)
                    file.Delete();
            }
            catch (Exception ex)
            {
                throw new StorageException(string.Format("Exception in FileSystemStorageProvider.RemoveItem({0})", file.FullName), ex);
            }
            finally
            {
                _sync.ReleaseWriterLock();
            }
        }

        #endregion

        #region Debug

        private static string _LogFilePath = string.Empty;
        private static bool _DebugMode = false;

        private static void Log(string message)
        {
            if (!_DebugMode || string.IsNullOrEmpty(_LogFilePath))
                return;
            try
            {
                string entry = string.Format("{0}\t{1}\r\n", DateTime.Now.ToString("dd/MM/yyyy HH:mm:ss"), message);
                File.AppendAllText(_LogFilePath, entry);
            }
            catch
            {
                return;
            }
        }

        #endregion

        #region Scavenge

        //
        // TODO: rivedere completamente il controllo del processo di scavenging, creando un nuovo Thread per ogni loop anzich un solo Thread in perenne esecuzione!
        //

        private Thread _scavengingProcess = null;
        private static readonly object _scavengingSync = new object();        

        public void StartScavenging()
        {
            if (!Settings.UseScavenging)
                return;
            lock (_scavengingSync)
            {
                if (_scavengingProcess == null || !_scavengingProcess.IsAlive)
                {
                    Log(string.Format("StartScavenging({0})", Settings.Id));
                    _scavengingProcess = new Thread(Scavenge);
                    _scavengingProcess.IsBackground = true;
                    _scavengingProcess.Priority = ThreadPriority.Lowest;
                    _scavengingProcess.Start();
                }
            }  
        }

        public void StopScavenging()
        {
            if (!Settings.UseScavenging)
                return;            
            lock (_scavengingSync)
            {
                if (_scavengingProcess != null && _scavengingProcess.IsAlive)
                {
                    _scavengingProcess.Abort();
                    _scavengingProcess.Join();
                    _scavengingProcess = null;
                }
            }
            Log(string.Format("StopScavenging({0})", Settings.Id));
        }

        private void Scavenge()
        {
            try
            {
                if (!Settings.UseScavenging)
                    return;
                Log(string.Format("Scavenge({0})", Settings.Id));
                List<FileInfo> fil = new List<FileInfo>();
                long storageSize = 0;
                DirectoryInfo root = new DirectoryInfo(Settings.Path);

                AnalyzeDirectory(root, ref fil, ref storageSize);

                Log(string.Format("{0} has {1} bytes in {2} items (max {3})", Settings.Id, storageSize.ToString("###,##0"), fil.Count, Settings.MaxStorageSizeBytes.ToString("###,##0")));

                fil.Sort(delegate(FileInfo f1, FileInfo f2)
                             {
                                 return f1.LastAccessTime.CompareTo(f2.LastAccessTime);
                             });

                if (Settings.MaxStorageSizeBytes > 0 && storageSize > Settings.MaxStorageSizeBytes && Settings.MinFreeSpaceAfterScavengingBytes > 0)
                {
                    int itemsRemoved = 0;
                    foreach (FileInfo f in fil)
                    {
                        try
                        {
                            long fs = f.Length;
                            RemoveItem(f);
                            storageSize -= fs;
                            itemsRemoved++;
                        }
                        catch (Exception ex)
                        {
                            Log(string.Format("ERR@Scavenge() deleting {0}: {1}", f.FullName, ex.Message));
                        }
                        if (storageSize <= Settings.MinFreeSpaceAfterScavengingBytes)
                            break;
                    }
                    Log(string.Format("{0} has now {1} bytes ({2} items removed)", Settings.Id, storageSize, itemsRemoved));
                }                
            }
            finally
            {
                Log(string.Format("Scavenge({0}) completed", Settings.Id));
                if (Settings.ScavengingInterval > TimeSpan.Zero)
                {
                    Log(string.Format("Sleeping({0}) for {1}", Settings.Id, Settings.ScavengingInterval));
                    Thread.Sleep(Settings.ScavengingInterval);
                }
                Scavenge();
            }
        }

        private void AnalyzeDirectory(DirectoryInfo directory, ref List<FileInfo> files, ref long size)
        {
            foreach (FileInfo f in directory.GetFiles("*" + FileExtension))
            {
                try
                {
                    if (GetItem(f).IsExpired())
                        RemoveItem(f);
                    else
                    {
                        files.Add(f);
                        size += f.Length;
                    }
                }
                catch(Exception ex)
                {
                    Log(string.Format("ERR@AnalyzeDirectory() scavenging {0}: {1}", f.FullName, ex.Message));
                }
            }
            
            foreach (DirectoryInfo d in directory.GetDirectories())
                AnalyzeDirectory(d, ref files, ref size);

            if (directory.GetFiles().Length == 0 && directory.GetDirectories().Length == 0)
            {
                try
                {
                    directory.Delete(false);
                }
                catch(Exception ex)
                {
                    Log(string.Format("ERR@AnalyzeDirectory() deleting {0}: {1}", directory.FullName, ex.Message));
                }
            }
        }
        
        #endregion

    }
}
