using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Text;
using IFXplus.SmartStorage.Providers;

namespace IFXplus.SmartStorage
{

    /// <summary>
    /// StorageItemMetadata
    /// </summary>
    [Serializable]
    public class StorageItemMetadata
    {

        #region Consts

        public static readonly TimeSpan NoAbsoluteTTL = TimeSpan.Zero;
        
        #endregion

        #region Fields

        private string _BaseKey = string.Empty;
        private StringDictionary _KeyParameters = new StringDictionary();
        private TimeSpan _AbsoluteTTL = NoAbsoluteTTL;
        private List<StorageProviderReference> _Providers = new List<StorageProviderReference>();
        
        #endregion

        #region Constructor
		
        /// <summary>
        /// Crea una nuova istanza di StorageItemMetadata
        /// </summary>
        /// <remarks>Non utilizzare il costruttore vuoto per istanziare una nuova istanza di StorageItemMetadata (ha visibilit "public" solo per consentire la serializzazione)</remarks>                
        public StorageItemMetadata() :
            this("") { }

        /// <summary>
        /// Crea una nuova istanza di StorageItemMetadata
        /// </summary>
        /// <param name="key">Key</param>
        public StorageItemMetadata(string key) : 
            this(key, new StringDictionary()) {}

        /// <summary>
        /// Crea una nuova istanza di StorageItemMetadata
        /// </summary>
        /// <param name="key">Key</param>
        /// <param name="keyParameters">KeyParameters</param>
        public StorageItemMetadata(string key, StringDictionary keyParameters) :
            this(key, keyParameters, NoAbsoluteTTL) { }

        /// <summary>
        /// Crea una nuova istanza di StorageItemMetadata
        /// </summary>
        /// <param name="key">Key</param>
        /// <param name="keyParameters">KeyParameters</param>
        /// <param name="absoluteTTL">AbsoluteTTL</param>
        public StorageItemMetadata(string key, StringDictionary keyParameters, TimeSpan absoluteTTL) :
            this(key, keyParameters, absoluteTTL, new List<StorageProviderReference>()) { }
        
        /// <summary>
        /// Crea una nuova istanza di StorageItemMetadata
        /// </summary>
        /// <param name="basekey">BaseKey</param>
        /// <param name="keyParameters">KeyParameters</param>
        /// <param name="absoluteTTL">AbsoluteTTL</param>
        /// <param name="providers">Providers</param>
        public StorageItemMetadata(string basekey, StringDictionary keyParameters, TimeSpan absoluteTTL, List<StorageProviderReference> providers)
        {
            BaseKey = basekey;
            KeyParameters = keyParameters;
            AbsoluteTTL = absoluteTTL;
            Providers = providers;
        }

        #endregion

        #region Properties

        /// <summary>
        /// BaseKey
        /// </summary>
        public string BaseKey
        {
            get { return _BaseKey; }
            set { _BaseKey = value; }
        }
        
        /// <summary>
        /// KeyParameters
        /// </summary>
        public StringDictionary KeyParameters
        {
            get { return _KeyParameters; }
            set { _KeyParameters = value; }  
        }

        /// <summary>
        /// AbsoluteTTL
        /// </summary>
        public TimeSpan AbsoluteTTL
        {
            get { return _AbsoluteTTL; }
            set { _AbsoluteTTL = value; }
        }        
        
        /// <summary>
        /// Providers
        /// </summary>
        public List<StorageProviderReference> Providers
        {
            get { return _Providers; }
            set { _Providers = value; }
        }
        
        #endregion

        #region Properties (calculated)

        /// <summary>
        /// Key calcolata dall'unione di BaseKey e KeyParameters
        /// </summary>
        public string Key
        {
            get
            {
                KeyParameters = KeyParameters ?? new StringDictionary();
                StringBuilder sb = new StringBuilder();
                sb.Append(BaseKey);
                sb.Append('(');
                bool isFirst = true;
                foreach (String k in KeyParameters.Keys)
                {
                    if (!isFirst)
                        sb.Append(',');
                    else
                        isFirst = false;
                    sb.Append(k + "=" + KeyParameters[k]);
                }
                sb.Append(')');
                return sb.ToString();
            }
        }

        #endregion

        #region Methods

        /// <summary>
        /// Aggiunge un parametro per la generazione della chiave
        /// </summary>
        /// <param name="key">Chiave da aggiungere</param>
        /// <param name="value">Valore da aggiungere</param>
        public void AddKeyParameter(string key, string value)
        {
            KeyParameters.Add(key, value);
        }

        /// <summary>
        /// Aggiunge un provider all'elenco dei providers da utilizzare per lo storage
        /// </summary>
        /// <param name="provider">Il provider da aggiungere</param>
        public void AddProvider(StorageProviderReference provider)
        {
            _Providers.Add(provider);
        }

        /// <summary>
        /// Aggiunge un provider all'elenco dei providers da utilizzare per lo storage
        /// </summary>
        /// <param name="id">Id del provider da aggiungere</param>
        /// <param name="type">Tipo di provider da aggiungere</param>
        public void AddProvider(string id, StorageProviderType type)
        {
            AddProvider(new StorageProviderReference(id, type));
        }

        /// <summary>
        /// Restituisce una stringa con i valori principali per l'identificazione dell'istanza
        /// </summary>
        /// <returns>La stringa che rappresenta l'istanza corrente</returns>
        public override string ToString()
        {
            StringBuilder spr = new StringBuilder();
            spr.Append('[');
            bool isFirst = true;
            foreach (StorageProviderReference provider in Providers)
            {
                if (!isFirst)
                    spr.Append(',');
                else
                    isFirst = false;
                spr.Append(provider.ToString());
            }
            spr.Append(']');
            return string.Format("{{ \"Key\" : \"{0}\", \"AbsoluteTTL\" : {1}, \"Providers\" : {{ {2} }} }}", Key, AbsoluteTTL, spr);
        }

        #endregion

    }
}